# auto generated by update_py.py
import json

import tlclient.trader.message_common as message_common
import tlclient.trader.message_market as message_market
import tlclient.trader.message_trade as message_trade
from tlclient.trader.constant import ExchangeID, OffsetFlag, OrderType, MsgType
from tlclient.linker.constant import FistType
from tlclient.linker.fist import Fist
from tlclient.linker.frame import Frame
from tlclient.linker.utility import bytify
from tlclient.linker.timer import Timer


class Client(Fist):

    def __init__(self, name, env_name, addr):
        Fist.__init__(self, name, FistType.CLIENT, env_name)
        self.set_master_addr(addr)
        self.create_fist()
        self.trade_router = None
        self.market_router = None
        self.oms_name = None

    def init_trade(self, router_name):
        self.reg_req(router_name)
        self.reg_sub(router_name)
        self.trade_router = router_name

    def init_market(self, router_name):
        self.reg_req(router_name)
        self.reg_sub(router_name)
        self.market_router = router_name

    def init_oms(self, oms_name):
        self.reg_req(oms_name)
        self.oms_name = oms_name

    def get_order_info(self, order_id):
        if self.oms_name is None:
            print('oms has not been inited!')
            return None
        req = message_trade.ReqOrderInfo()
        req.order_id = order_id
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.oms_name, req, MsgType.REQ_ORDER_INFO, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            order_info = ret.get_obj(message_trade.RspOrderInfo)
            return order_info
        else:
            print('[get_order_info] error occurred (eid){}'.format(err_id))
            return None

    def get_position(self, tg_name):
        if self.oms_name is None:
            print('oms has not been inited!')
            return None
        req = message_trade.ReqPosition()
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.oms_name, req, MsgType.REQ_POSITION, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            pos_d = json.loads(ret.get_string())
            pos_obj = message_trade.RspPosition(pos_d)
            return pos_obj
        else:
            print('[get_position] error occurred (eid){}'.format(err_id))
            return None

    def get_account(self, tg_name):
        if self.oms_name is None:
            print('oms has not been inited!')
            return None
        req = message_trade.ReqAccount()
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.oms_name, req, MsgType.REQ_ACCOUNT, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            acc_obj = ret.get_obj(message_trade.RspAccount)
            return acc_obj
        else:
            print('[get_account] error occurred (eid){}'.format(err_id))
            return None

    # return order_id
    def insert_order(self, tg_name, exchange, ticker, price, volume, order_type, direction, offset_flag=OffsetFlag.NOT_AVAILABLE):
        req = message_trade.ReqOrderInsert()
        req.tg_name = bytify(tg_name)
        req.exchange = exchange
        req.ticker = bytify(ticker)
        req.price = price
        req.volume = volume
        req.order_type = order_type
        req.direction = direction
        req.offset_flag = offset_flag
        req.client_name = bytify(self.fist_name)
        req.parent_id = -1
        ret = self.req(self.trade_router, req, MsgType.REQ_ORDER_INSERT, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[order_insert] error occurred (eid){}'.format(err_id))
            return -1

    def insert_twap_order(self, tg_name, exchange, ticker, volume, direction, start_time, end_time, sec_interval):
        req = message_trade.ReqTwapInsert()
        req.tg_name = bytify(tg_name)
        req.exchange = exchange
        req.ticker = bytify(ticker)
        req.price = 0
        req.volume = volume
        req.order_type = OrderType.TWAP
        req.direction = direction
        req.offset_flag = OffsetFlag.NOT_AVAILABLE
        req.client_name = bytify(self.fist_name)
        req.start_nano = Timer.get_nano(start_time)
        req.end_nano = Timer.get_nano(end_time)
        req.sec_interval = sec_interval
        req.parent_id = -1
        ret = self.req(self.trade_router, req, MsgType.REQ_ORDER_INSERT, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[insert_twap_order] error occurred (eid){}'.format(err_id))
            return -1

    def insert_basket_order(self, tg_name, basket_order, child_order_type, basket_name='', **algo_args):
        basket_order.basket_name = basket_name
        basket_order.tg_name = tg_name
        basket_order.order_id = -1
        basket_order.client_name = self.fist_name
        basket_order.set_child_order_type(child_order_type)
        basket_order.set_algo_info(**algo_args)
        basket_order_str = json.dumps(basket_order.to_dict())
        ret = self.req(self.trade_router, basket_order_str, MsgType.REQ_ORDER_INSERT_BASKET, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[insert_basket_order] error occurred (eid){}'.format(err_id))
            return -1

    def insert_basket_order_w_twap(self, tg_name, basket_order, start_time, end_time, sec_interval, basket_name=''):
        start_nano = Timer.get_nano(start_time)
        end_nano = Timer.get_nano(end_time)
        return self.insert_basket_order(tg_name, basket_order, OrderType.TWAP, basket_name=basket_name, start_nano=start_nano, end_nano=end_nano, sec_interval=sec_interval)

    def cancel_order(self, order_id, tg_name):
        req = message_trade.ReqOrderCancel()
        req.order_id = order_id
        req.tg_name = bytify(tg_name)

        ret = self.req(self.trade_router, req, MsgType.REQ_ORDER_CANCEL, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[cancel_order] error occurred (eid){}'.format(err_id))
            return -1

    def req_position(self, tg_name):
        req = message_trade.ReqPosition()
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.trade_router, req, MsgType.REQ_POSITION, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[req_position] error occurred (eid){} (tg){}'.format(err_id, tg_name))
            return -1

    def req_account(self, tg_name):
        req = message_trade.ReqAccount()
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.trade_router, req, MsgType.REQ_ACCOUNT, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[req_account] error occurred (eid){} (tg){}'.format(err_id, tg_name))
            return -1

    def req_active_orders(self, tg_name):
        req = message_trade.ReqActiveOrders()
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.trade_router, req, MsgType.REQ_ACTIVE_ORDERS, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[req_active_orders] error occurred (eid){} (tg){}'.format(err_id, tg_name))
            return -1

    def req_cancel_active_orders(self, tg_name, req_active_orders_id=-1):
        req = message_trade.ReqCancelActiveOrders()
        req.req_active_orders_id = req_active_orders_id
        req.tg_name = bytify(tg_name)
        req.client_name = bytify(self.fist_name)
        ret = self.req(self.trade_router, req, MsgType.REQ_CANCEL_ACTIVE_ORDERS, 0)
        err_id = ret.get_err_id()
        if err_id == 0:
            return ret.get_req_id()
        else:
            print('[req_cancel_active_orders] error occurred (eid){} (tg){}'.format(err_id, tg_name))
            return -1

    # subscription
    def subscribe_snap(self, exchange, ticker):
        return self.subscribe(exchange, ticker, MsgType.MKT_SNAP)

    def subscribe_bar(self, exchange, ticker):
        return self.subscribe(exchange, ticker, MsgType.MKT_BAR)

    def subscribe_index(self, exchange, ticker):
        return self.subscribe(exchange, ticker, MsgType.MKT_INDEX)

    def subscribe_order(self, exchange, ticker):
        return self.subscribe(exchange, ticker, MsgType.MKT_ORDER)

    def subscribe_trade(self, exchange, ticker):
        return self.subscribe(exchange, ticker, MsgType.MKT_TRADE)

    def subscribe(self, exchange, ticker, msg_type):
        return self.subscribe_batch([exchange], [ticker], msg_type)

    def subscribe_batch(self, exchanges, tickers, msg_type=MsgType.MKT_SNAP):
        if not isinstance(tickers, list):
            tickers = [tickers]
        if not isinstance(exchanges, list):
            exchanges = [exchanges for i in range(0, len(tickers))]
        if len(exchanges) != len(tickers):
            print("[subscribe] exchanges and tickers, size mismatch")
            return False
        ss = json.dumps([{'exchange': ExchangeID.read(exchanges[i]), 'ticker': tickers[i], 'msg_type': msg_type} for i in range(0, len(tickers))])
        rsp = self.req(self.market_router, ss, MsgType.MKT_SUBSCRIBE, 0)
        if rsp.get_err_id() != 0:
            print("[subscribe] failed (err_id){} (msg){}".format(rsp.get_err_id(), ss))
            return False
        else:
            print('[subscribe] success')
            return True

    # to override
    # trade
    def on_rsp_order_insert(self, obj, frame_nano):
        pass

    def on_rsp_order_cancel(self, obj, frame_nano):
        pass

    def on_rtn_order(self, obj, frame_nano):
        pass

    def on_rtn_trade(self, obj, frame_nano):
        pass

    def on_rsp_position(self, obj, frame_nano):
        pass

    def on_rsp_account(self, obj, frame_nano):
        pass

    def on_rsp_active_orders(self, obj, frame_nano):
        pass

    def on_rsp_cancel_active_orders(self, obj, frame_nano):
        pass

    # to override
    # market
    def on_mkt_snap(self, obj, msg_type, frame_nano):
        pass

    def on_mkt_bar(self, obj, msg_type, frame_nano):
        pass

    def on_mkt_index(self, obj, msg_type, frame_nano):
        pass

    def on_mkt_order(self, obj, msg_type, frame_nano):
        pass

    def on_mkt_trade(self, obj, msg_type, frame_nano):
        pass

    # to override
    # system status
    def on_gateway_connection_change(self, obj, frame_nano):
        pass

    def on_gateway_heart_beat(self, obj, frame_nano):
        pass

    ######################
    # internal functions #
    ######################

    def on_pub_frame(self, f):
        msg_type = f.get_msg_type()
        frame_nano = f.get_nano()
        if MsgType.is_market_data_type(msg_type):
            if msg_type == MsgType.MKT_SNAP:
                snap_obj = f.get_obj(message_market.MktSnap)
                self.on_mkt_snap(snap_obj, msg_type, frame_nano)

            elif msg_type == MsgType.MKT_SNAP_PLUS:
                snap_obj = f.get_obj(message_market.MktSnapPlus)
                self.on_mkt_snap(snap_obj, msg_type, frame_nano)

            elif msg_type == MsgType.MKT_SNAP_FUT:
                snap_obj = f.get_obj(message_market.MktSnapFut)
                self.on_mkt_snap(snap_obj, msg_type, frame_nano)

            elif msg_type == MsgType.MKT_SNAP_OPT:
                snap_obj = f.get_obj(message_market.MktSnapOpt)
                self.on_mkt_snap(snap_obj, msg_type, frame_nano)

            elif msg_type == MsgType.MKT_BAR:
                bar_obj = f.get_obj(message_market.MktBar)
                self.on_mkt_bar(bar_obj, msg_type, frame_nano)

            elif msg_type == MsgType.MKT_BAR_GEN:
                bar_obj = f.get_obj(message_market.MktBarGen)
                self.on_mkt_bar(bar_obj, msg_type, frame_nano)

        elif MsgType.is_trading_data_type(msg_type):
            if msg_type == MsgType.RSP_ORDER_INSERT:
                rsp_ord_obj = f.get_obj(message_trade.RspOrderInsert)
                self.on_rsp_order_insert(rsp_ord_obj, frame_nano)

            elif msg_type == MsgType.RSP_ORDER_CANCEL:
                rsp_cancel_obj = f.get_obj(message_trade.RspOrderCancel)
                self.on_rsp_order_cancel(rsp_cancel_obj, frame_nano)

            elif msg_type == MsgType.RTN_ORDER:
                ord_object = f.get_obj(message_trade.RtnOrder)
                self.on_rtn_order(ord_object, frame_nano)

            elif msg_type == MsgType.RTN_TRADE:
                trd_obj = f.get_obj(message_trade.RtnTrade)
                self.on_rtn_trade(trd_obj, frame_nano)

            elif msg_type == MsgType.RSP_POSITION:
                pos_d = json.loads(f.get_string())
                pos_obj = message_trade.RspPosition(pos_d)
                self.on_rsp_position(pos_obj, frame_nano)

            elif msg_type == MsgType.RSP_ACCOUNT:
                acc_obj = f.get_obj(message_trade.RspAccount)
                self.on_rsp_account(acc_obj, frame_nano)

            elif msg_type == MsgType.RSP_ACTIVE_ORDERS:
                ods_d = json.loads(f.get_string())
                ods_obj = message_trade.RspActiveOrders(ods_d)
                self.on_rsp_active_orders(ods_obj, frame_nano)

            elif msg_type == MsgType.RSP_CANCEL_ACTIVE_ORDERS:
                rsp_obj = f.get_obj(message_trade.RspCancelActiveOrders)
                self.on_rsp_cancel_active_orders(rsp_obj, frame_nano)

        elif MsgType.is_system_status_data_type(msg_type):
            if msg_type == MsgType.GTW_CONNECTION:
                gtw_con_obj = f.get_obj(message_common.GatewayConnectionStatus)
                self.on_gateway_connection_change(gtw_con_obj, frame_nano)

            elif msg_type == MsgType.GTW_HEART_BEAT:
                gtw_hb = f.get_obj(message_common.GatewayHeartBeat)
                self.on_gateway_heart_beat(gtw_hb, frame_nano)
